@@ -6,14 +6,16 @@ Originally (loosely) based on code from py2exe's build_exe.py by Thomas Heller.
 from pkg_resources import require
 require("altgraph", "modulegraph", "macholib")
 
+import imp
 import sys
 import os
 import zipfile
 import plistlib
 import shlex
+import shutil
 from cStringIO import StringIO
 
-from distutils.core import Command
+from setuptools import Command
 from distutils.util import convert_path
 from distutils import log
 from distutils.errors import *
@@ -21,7 +23,7 @@ from distutils.errors import *
 from altgraph.compat import *
 
 from modulegraph.find_modules import find_modules, parse_mf_results
-from modulegraph.modulegraph import SourceModule
+from modulegraph.modulegraph import SourceModule, Package, os_listdir
 
 import macholib.dyld
 import macholib.MachOStandalone
@@ -31,11 +33,13 @@ from py2app.create_pluginbundle import create_pluginbundle
 from py2app.util import \
     fancy_split, byte_compile, make_loader, imp_find_module, \
     copy_tree, fsencoding, strip_files, in_system_path, makedirs, \
-    iter_platform_files, find_version, skipscm, momc
+    iter_platform_files, find_version, skipscm, momc, copy_file, \
+    os_path_isdir, copy_resource
 from py2app.filters import \
     not_stdlib_filter, not_system_filter, has_filename_filter
 from py2app import recipes
 
+
 def get_zipfile(dist):
     return getattr(dist, "zipfile", None) or "site-packages.zip"
 
@@ -105,6 +109,10 @@ def validate_target(dist, attr, value):
 def FixupTargets(targets, default_attribute):
     if not targets:
         return targets
+    try:
+        targets = eval(targets)
+    except:
+        pass
     ret = []
     for target_def in targets:
         if isinstance(target_def, basestring):
@@ -145,8 +153,11 @@ class py2app(Command):
     # List of option tuples: long name, short name (None if no short
     # name), and help string.
 
-
     user_options = [
+        ("app=", None,
+         "application bundle to be built"),
+        ("plugin=", None,
+         "puglin bundle to be built"),
         ('optimize=', 'O',
          "optimization level: -O1 for \"python -O\", "
          "-O2 for \"python -OO\", and -O0 to disable [default: -O0]"),
@@ -162,6 +173,8 @@ class py2app(Command):
          "comma-separated list of frameworks or dylibs to exclude"),
         ("datamodels=", None,
          "xcdatamodels to be compiled and copied into Resources"),
+        ("mappingmodels=", None,
+          "xcmappingmodels to be compiled and copied into Resources"),
         ("resources=", 'r',
          "comma-separated list of additional data files and folders to include (not for code!)"),
         ("frameworks=", 'f',
@@ -198,10 +211,12 @@ class py2app(Command):
          "directory to put final built distributions in (default is dist)"),
         ('site-packages', None,
          "include the system and user site-packages into sys.path"),
-        ('debug-modulegraph', None,
-         'Drop to pdb console after the module finding phase is complete'),
         ("strip", 'S',
          "strip debug and local symbols from output (on by default, for compatibility)"),
+        ("prefer-ppc", None,
+         "Force application to run translated on i386 (LSPrefersPPC=True)"),
+        ('debug-modulegraph', None,
+         'Drop to pdb console after the module finding phase is complete'),
         ("debug-skip-macholib", None,
          "skip macholib phase (app will not be standalone!)"),
         ]
@@ -221,9 +236,12 @@ class py2app(Command):
         "debug-modulegraph",
         "debug-skip-macholib",
         "graph",
+        "prefer-ppc",
     ]
 
     def initialize_options (self):
+        self.app = None
+        self.plugin = None
         self.bdist_base = None
         self.xref = False
         self.graph = False
@@ -246,12 +264,14 @@ class py2app(Command):
         self.frameworks = None
         self.resources = None
         self.datamodels = None
+        self.mappingmodels = None
         self.plist = None
         self.compressed = True
         self.semi_standalone = is_system()
         self.dist_dir = None
         self.debug_skip_macholib = False
         self.debug_modulegraph = False
+        self.prefer_ppc = False
         self.filters = []
         self.eggs = []
 
@@ -321,6 +341,13 @@ class py2app(Command):
         self.runtime_preferences = list(self.get_runtime_preferences())
 
 
+        if self.datamodels:
+            print "WARNING: the datamodels option is deprecated, add model files to the list of resources"
+
+        if self.mappingmodels:
+            print "WARNING: the mappingmodels option is deprecated, add model files to the list of resources"
+
+
     def get_default_plist(self):
         # XXX - this is all single target stuff
         plist = {}
@@ -422,8 +449,26 @@ class py2app(Command):
 
     def compile_datamodels(self, resdir):
         for src, dest in self.iter_datamodels(resdir):
+            print "compile datamodel", src, "->", dest
             self.mkpath(os.path.dirname(dest))
             momc(src, dest)
+
+    def iter_mappingmodels(self, resdir):
+        for (path, files) in imap(normalize_data_file, self.mappingmodels or ()):
+            path = fsencoding(path)
+            for fn in files:
+                fn = fsencoding(fn)
+                basefn, ext = os.path.splitext(fn)
+                if ext != '.xcmappingmodel':
+                    basefn = fn
+                    fn += '.xcmappingmodel'
+                destfn = os.path.basename(basefn) + '.cdm'
+                yield fn, os.path.join(resdir, path, destfn)
+
+    def compile_mappingmodels(self, resdir):
+        for src, dest in self.iter_mappingmodels(resdir):
+            self.mkpath(os.path.dirname(dest))
+            mapc(src, dest)
         
     def iter_data_files(self):
         dist = self.distribution
@@ -455,6 +500,7 @@ class py2app(Command):
                 argv_emulation=bool(self.argv_emulation),
                 no_chdir=bool(self.no_chdir),
                 optimize=self.optimize,
+                prefer_ppc=self.prefer_ppc,
             ),
         )
 
@@ -465,6 +511,19 @@ class py2app(Command):
             plist.update(getattr(target, 'plist', {}))
         plist.update(self.plist)
         plist.update(self.get_plist_options())
+
+        if self.iconfile:
+            iconfile = self.iconfile
+            if not os.path.exists(iconfile):
+                iconfile = iconfile + '.icns'
+            if not os.path.exists(iconfile):
+                raise DistutilsOptionError("icon file must exist: %r"
+                    % (self.iconfile,))
+            self.resources.append(iconfile)
+            plist[u'CFBundleIconFile'] = os.path.basename(iconfile)
+        if self.prefer_ppc:
+            plist[u'LSPrefersPPC'] = True
+
         self.plist = plist
         return plist
 
@@ -664,6 +723,10 @@ class py2app(Command):
                      verbose=self.verbose,
                      dry_run=self.dry_run)
 
+        for item in py_files:
+            if not isinstance(item, Package): continue
+            self.copy_package_data(item, self.collect_dir)
+
         self.lib_files = []
         self.app_files = []
 
@@ -671,10 +734,15 @@ class py2app(Command):
         archive_name = os.path.join(self.lib_dir,
                                     os.path.basename(get_zipfile(dist)))
 
-        for (path, files) in loader_files:
+        for path, files in loader_files:
             dest = os.path.join(self.collect_dir, path)
+            self.mkpath(dest)
             for fn in files:
-                self.copy_file(fn, os.path.join(dest, os.path.basename(fn)))
+                destfn = os.path.join(dest, os.path.basename(fn))
+                if os.path.isdir(fn):
+                    self.copy_tree(fn, destfn, preserve_symlinks=False)
+                else:
+                    self.copy_file(fn, destfn)
 
         arcname = self.make_lib_archive(archive_name,
             base_dir=self.collect_dir, verbose=self.verbose,
@@ -687,7 +755,7 @@ class py2app(Command):
                 target, arcname, pkgexts, copyexts, target.script)
             exp = os.path.join(dst, 'Contents', 'MacOS')
             execdst = os.path.join(exp, 'python')
-            if is_system() or self.semi_standalone:
+            if self.semi_standalone:
                 self.symlink(sys.executable, execdst)
             else:
                 self.copy_file(sys.executable, execdst)
@@ -708,9 +776,66 @@ class py2app(Command):
                     mm.mm.run_file(fmwk)
                 platfiles = mm.run()
                 if self.strip:
+                    platfiles = self.strip_dsym(platfiles)
                     self.strip_files(platfiles)
             self.app_files.append(dst)
 
+    def copy_package_data(self, package, target_dir):
+        """
+        Copy any package data in a python package into the target_dir.
+
+        This is a bit of a hack, it would be better to identify python eggs
+        and copy those in whole.
+        """
+        exts = [ i[0] for i in imp.get_suffixes() ]
+        exts.append('.py')
+        exts.append('.pyc')
+        exts.append('.pyo')
+        def datafilter(item):
+            for e in exts:
+                if item.endswith(e):
+                    return False
+            return True
+
+        target_dir = os.path.join(target_dir, *(package.identifier.split('.')))
+        for dname in package.packagepath:
+            filenames = filter(datafilter, os_listdir(dname))
+            for fname in filenames:
+                if fname in ('.svn', 'CVS'):
+                    # Scrub revision manager junk
+                    continue
+                pth = os.path.join(dname, fname)
+
+                # Check if we have found a package, exclude those
+                if os_path_isdir(pth):
+                    for p in os_listdir(pth):
+                        if p.startswith('__init__.') and p[8:] in exts:
+                            break
+
+                    else:
+                        copy_tree(pth, os.path.join(target_dir, fname))
+                    continue
+                else:
+                    copy_file(pth, os.path.join(target_dir, fname))
+
+
+    def strip_dsym(self, platfiles):
+        """ Remove .dSYM directories in the bundled application """
+
+        #
+        # .dSYM directories are contain detached debugging information and
+        # should be completely removed when the "strip" option is specified.
+        #
+        if self.dry_run:
+            return platfiles
+        for dirpath, dnames, fnames in os.walk(self.appdir):
+            for nm in list(dnames):
+                if nm.endswith('.dSYM'):
+                    print "removing debug info: %s/%s"%(dirpath, nm)
+                    shutil.rmtree(os.path.join(dirpath, nm))
+                    dnames.remove(nm)
+        return [file for file in platfiles if '.dSYM' not in file]
+
     def strip_files(self, files):
         unstripped = 0L
         stripfiles = []
@@ -783,22 +908,44 @@ class py2app(Command):
         indir = os.path.dirname(os.path.join(info['location'], info['name']))
         outdir = os.path.dirname(os.path.join(dst, info['name']))
         self.mkpath(os.path.join(outdir, 'Resources'))
+        pydir = 'python%d.%d'%(info['version'])
+
+        # distutils looks for some files relative to sys.executable, which
+        # means they have to be in the framework...
+        self.mkpath(os.path.join(outdir, 'include'))
+        self.mkpath(os.path.join(outdir, 'include', pydir))
+        self.mkpath(os.path.join(outdir, 'lib'))
+        self.mkpath(os.path.join(outdir, 'lib', pydir))
+        self.mkpath(os.path.join(outdir, 'lib', pydir, 'config'))
         fmwkfiles = [
             os.path.basename(info['name']),
             'Resources/Info.plist',
-            'Resources/version.plist',
+            'include/%s/pyconfig.h'%(pydir,),
+            'lib/%s/config/Makefile'%(pydir,)
         ]
         for fn in fmwkfiles:
             self.copy_file(
                 os.path.join(indir, fn),
                 os.path.join(outdir, fn))
 
+
+
     def fixup_distribution(self):
         dist = self.distribution
 
+        # Trying to obtain app and plugin from dist for backward compatibility
+        # reasons.
+        app = dist.app
+        plugin = dist.plugin
+        # If we can get suitable values from self.app and self.plugin, we prefer
+        # them.
+        if self.app is not None or self.plugin is not None:
+            app = self.app
+            plugin = self.plugin
+
         # Convert our args into target objects.
-        dist.app = FixupTargets(dist.app, "script")
-        dist.plugin = FixupTargets(dist.plugin, "script")
+        dist.app = FixupTargets(app, "script")
+        dist.plugin = FixupTargets(plugin, "script")
         if dist.app and dist.plugin:
             # XXX - support apps and plugins?
             raise DistutilsOptionError(
@@ -845,15 +992,9 @@ class py2app(Command):
         prescripts = []
         if self.site_packages or self.alias:
             prescripts.append('site_packages')
-        if self.iconfile:
-            iconfile = self.iconfile
-            if not os.path.exists(iconfile):
-                iconfile = iconfile + '.icns'
-            if not os.path.exists(iconfile):
-                raise DistutilsOptionError("icon file must exist: %r"
-                    % (self.icon,))
-            self.resources.append(iconfile)
-            self.plist[u'CFBundleIconFile'] = os.path.basename(iconfile)
+
+        #if self.style == 'app':
+        #    prescripts.append('setup_pkgresource')
 
         if self.argv_emulation and self.style == 'app':
             prescripts.append('argv_emulation')
@@ -965,7 +1106,7 @@ class py2app(Command):
             if fmwk is None:
                 yield fn
             else:
-                basename = fmwk['name'] + '.framework'
+                basename = fmwk['shortname'] + '.framework'
                 yield os.path.join(fmwk['location'], basename)
     
     def build_alias_executable(self, target, script):
@@ -987,6 +1128,7 @@ class py2app(Command):
             
         
         # symlink data files
+        # XXX: fixme: need to integrate automatic data conversion
         for src, dest in self.iter_data_files():
             dest = os.path.join(resdir, dest)
             if src == dest:
@@ -1004,6 +1146,7 @@ class py2app(Command):
             self.symlink(os.path.abspath(src), dest)
 
         self.compile_datamodels(resdir)
+        self.compile_mappingmodels(resdir)
 
         from Carbon.File import FSRef
         aliasdata = FSRef(script).FSNewAliasMinimal().data
@@ -1027,16 +1170,17 @@ class py2app(Command):
     def build_executable(self, target, arcname, pkgexts, copyexts, script):
         # Build an executable for the target
         appdir, resdir, plist = self.create_bundle(target, script)
+        self.appdir = appdir
+        self.resdir = resdir
+        self.plist = plist
 
         for src, dest in self.iter_data_files():
             dest = os.path.join(resdir, dest)
             self.mkpath(os.path.dirname(dest))
-            if os.path.isdir(src):
-                self.copy_tree(src, dest, preserve_symlinks=True)
-            else:
-                self.copy_file(src, dest)
+            copy_resource(src, dest, dry_run=self.dry_run)
 
         self.compile_datamodels(resdir)
+        self.compile_mappingmodels(resdir)
 
         bootfn = '__boot__'
         bootfile = file(os.path.join(resdir, bootfn + '.py'), 'w')
@@ -1053,8 +1197,10 @@ class py2app(Command):
         self.symlink('../../site.py', os.path.join(pydir, 'site.py'))
         cfgdir = os.path.join(pydir, 'config')
         realcfg = os.path.join(realhome, 'config')
+        real_include = os.path.join(sys.prefix, 'include')
         if self.semi_standalone:
             self.symlink(realcfg, cfgdir)
+            self.symlink(real_include, os.path.join(resdir, 'include'))
         else:
             self.mkpath(cfgdir)
             for fn in 'Makefile', 'Setup', 'Setup.local', 'Setup.config':
@@ -1062,19 +1208,21 @@ class py2app(Command):
                 if os.path.exists(rfn):
                     self.copy_file(rfn, os.path.join(cfgdir, fn))
 
+            inc_dir = os.path.join(resdir, 'include', 'python' + sys.version[:3])
+            self.mkpath(inc_dir)
+            self.copy_file(os.path.join(real_include, 'python%s/pyconfig.h'%(
+                sys.version[:3])), os.path.join(inc_dir, 'pyconfig.h'))
+
+
         self.copy_file(arcname, pydir)
         ext_dir = os.path.join(pydir, os.path.basename(self.ext_dir))
         self.copy_tree(self.ext_dir, ext_dir, preserve_symlinks=True)
         self.copy_tree(self.framework_dir,
             os.path.join(appdir, 'Contents', 'Frameworks'),
             preserve_symlinks=True)
-        sitepkg = os.path.join(
-            pydir,
-            os.path.splitext(os.path.basename(arcname))[0],
-        )
         for pkg in self.packages:
             pkg = self.get_bootstrap(pkg)
-            dst = os.path.join(sitepkg, os.path.basename(pkg))
+            dst = os.path.join(pydir, os.path.basename(pkg))
             self.mkpath(dst)
             self.copy_tree(pkg, dst)
         for copyext in copyexts:
@@ -1083,7 +1231,7 @@ class py2app(Command):
                 os.path.splitext(copyext.filename)[1])
             )
             self.mkpath(os.path.dirname(fn))
-            self.copy_file(copyext.filename, fn)
+            copy_file(copyext.filename, fn, dry_run=self.dry_run)
 
         target.appdir = appdir
         return appdir
